package me.totalfreedom.discord;

import com.google.common.collect.ImmutableList;
import discord4j.common.util.Snowflake;
import discord4j.core.object.entity.Guild;
import discord4j.core.object.entity.Member;
import discord4j.core.object.entity.Message;
import discord4j.core.object.entity.Role;
import discord4j.core.object.entity.channel.TextChannel;
import discord4j.core.object.reaction.ReactionEmoji;
import discord4j.core.spec.EmbedCreateSpec;
import discord4j.core.spec.MessageCreateSpec;
import me.totalfreedom.discord.util.SnowflakeEntry;
import me.totalfreedom.discord.util.Utilities;
import me.totalfreedom.totalfreedommod.admin.Admin;
import me.totalfreedom.totalfreedommod.api.TFD4JCommons;
import me.totalfreedom.totalfreedommod.config.ConfigEntry;
import me.totalfreedom.totalfreedommod.player.PlayerData;
import me.totalfreedom.totalfreedommod.rank.Rank;
import me.totalfreedom.totalfreedommod.util.FLog;
import org.apache.commons.lang.RandomStringUtils;
import org.apache.commons.lang.WordUtils;
import org.bukkit.OfflinePlayer;
import org.bukkit.entity.Player;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;

public class TFD4JCommonsImpl implements TFD4JCommons
{
    private final TFD4J tfd4J;
    private final ImmutableList<String> DISCORD_SUBDOMAINS;
    private Flux<Message> sentMessages;

    public TFD4JCommonsImpl(TFD4J tfd4J)
    {
        this.tfd4J = tfd4J;
        this.sentMessages = Flux.fromIterable(new ArrayList<>());
        this.DISCORD_SUBDOMAINS = (ImmutableList<String>)
                List.of("discordapp.com",
                        "discord.com",
                        "discord.gg");
    }

    @Override
    public void messageAdminChatChannel(String message)
    {
        String chat_channel_id = ConfigEntry.DISCORD_ADMINCHAT_CHANNEL_ID.getString();

        String sanitizedMessage = sanitizeChatMessage(message);

        if (sanitizedMessage.isBlank()) return;

        if (!chat_channel_id.isEmpty())
        {
            MessageCreateSpec spec = MessageCreateSpec.builder()
                    .content(sanitizedMessage)
                    .build();

            Mono<Message> sentMessage = tfd4J
                    .getBot()
                    .getClient()
                    .getChannelById(SnowflakeEntry.adminChatChannelID)
                    .ofType(TextChannel.class)
                    .flatMap(c -> c.createMessage(spec));

            insert(sentMessage);
        }
    }

    public Flux<Message> getMessagesSent()
    {
        return sentMessages;
    }

    private void insert(Mono<Message> messageMono)
    {
        sentMessages.concatWith(messageMono);
    }

    @Override
    public void clearQueue()
    {
        sentMessages = Flux.fromIterable(new ArrayList<>());
    }

    @Override
    public void messageChatChannel(String message, boolean system)
    {
        String chat_channel_id = ConfigEntry.DISCORD_CHAT_CHANNEL_ID.getString();

        String sanitizedMessage = (system) ? message : sanitizeChatMessage(message);

        if (sanitizedMessage.isBlank()) return;

        if (!chat_channel_id.isEmpty())
        {
            MessageCreateSpec spec = MessageCreateSpec.builder()
                    .content(sanitizedMessage)
                    .build();

            Mono<Message> sentMessage = tfd4J
                    .getBot()
                    .getClient()
                    .getChannelById(SnowflakeEntry.chatChannelID)
                    .ofType(TextChannel.class)
                    .flatMap(c -> c.createMessage(spec));

            insert(sentMessage);
        }
    }

    private String sanitizeChatMessage(String message)
    {
        String newMessage = message;

        if (message.contains("@"))
        {
            // \u200B is Zero Width Space, invisible on Discord
            newMessage = message.replace("@", "@\u200B");
        }

        if (message.toLowerCase().contains("discord.gg")) // discord.gg/invite works as an invite
        {
            return "";
        }

        for (String subdomain : DISCORD_SUBDOMAINS)
        {
            if (message.toLowerCase().contains(subdomain + "/invite"))
            {
                return "";
            }
        }

        if (message.contains("§"))
        {
            newMessage = message.replace("§", "");
        }

        return Utilities.deformat(newMessage);
    }

    @Override
    public boolean syncRoles(Admin admin, String discordID)
    {
        if (discordID == null)
        {
            return false;
        }

        Guild server = tfd4J.getBot().getGuildById()
                .blockOptional()
                .orElseThrow();

        Member member = server.getMemberById(Snowflake.of(discordID))
                .blockOptional()
                .orElseThrow();

        Role adminRole = server.getRoleById(SnowflakeEntry.adminRoleID)
                .blockOptional()
                .orElseThrow();

        Role senioradminRole = server.getRoleById(SnowflakeEntry.seniorRoleID)
                .blockOptional()
                .orElseThrow();

        if (!admin.isActive())
        {
            syncRolesActivityCheck(member, adminRole, senioradminRole);
            return true;
        }

        if (admin.getRank().equals(Rank.ADMIN))
        {
            syncRolesAdminAssignment(member, adminRole, senioradminRole);
            return true;
        } else if (admin.getRank().equals(Rank.SENIOR_ADMIN))
        {
            syncRolesSeniorAssignment(member, adminRole, senioradminRole);
            return true;
        }
        return false;
    }

    private void syncRolesAdminAssignment(Member member, Role adminRole, Role senioradminRole)
    {
        member.getRoles().doFirst(() ->
        {
            if (!member.getRoles().collectList().blockOptional().orElseThrow().contains(adminRole))
            {
                member.addRole(adminRole.getId()).block();
            }
        }).doOnEach(r ->
        {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(senioradminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }

    private void syncRolesActivityCheck(Member member, Role adminRole, Role senioradminRole)
    {
        member.getRoles().doOnEach(r ->
        {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(adminRole) || role.equals(senioradminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }

    private void syncRolesSeniorAssignment(Member member, Role adminRole, Role senioradminRole)
    {
        member.getRoles().doFirst(() ->
        {
            if (!member.getRoles().collectList().blockOptional().orElseThrow().contains(senioradminRole))
            {
                member.addRole(senioradminRole.getId()).block();
            }
        }).doOnEach(r ->
        {
            Role role = r.get();
            if (role == null) return;

            if (role.equals(adminRole))
            {
                member.removeRole(role.getId()).block();
            }
        }).subscribe();
    }

    @Override
    public String getCode(PlayerData playerData)
    {
        for (String code : this.getLinkCodes().keySet())
        {
            if (this.getLinkCodes().get(code).equals(playerData))
            {
                return code;
            }
        }
        return "";
    }

    @Override
    public String generateCode(int size)
    {
        return RandomStringUtils.randomNumeric(size);
    }

    @Override
    public Map<String, PlayerData> getLinkCodes()
    {
        return tfd4J.getBot().getLinkCodes();
    }

    @Override
    public String formatBotTag()
    {
        return tfd4J.getBot().formatBotTag();
    }

    @Override
    public boolean sendReportOffline(Player reporter, OfflinePlayer reported, String reason)
    {
        if (!tfd4J.getBot().shouldISendReport())
        {
            return false;
        }

        final Guild server = tfd4J.getBot().getGuildById().block();

        if (server == null) return false;

        final TextChannel channel = server.getChannelById(SnowflakeEntry.reportChannelID)
                .ofType(TextChannel.class)
                .blockOptional()
                .orElseThrow();

        final EmbedCreateSpec.Builder builder = EmbedCreateSpec.builder()
                .title("Report for " + reported.getName() + " (offline)")
                .description(reason)
                .footer("Reported by " + reporter.getName(), "https://minotar.net/helm/" + reporter.getName() + ".png")
                .timestamp(Instant.from(ZonedDateTime.now()));
        if (tfd4J.getBot().getTFM().getCommons().esb.isEnabled())
        {
            com.earth2me.essentials.User user = tfd4J.getBot().getTFM().getCommons().esb.getEssentialsUser(reported.getName());
            String location = "World: " + Objects.requireNonNull(user.getLastLocation().getWorld()).getName() + ", X: " + user.getLastLocation().getBlockX() + ", Y: " + user.getLastLocation().getBlockY() + ", Z: " + user.getLastLocation().getBlockZ();
            builder.addField("Location", location, true);
            builder.addField("God Mode", WordUtils.capitalizeFully(String.valueOf(user.isGodModeEnabled())), true);
            if (user.getNickname() != null)
            {
                builder.addField("Nickname", user.getNickname(), true);
            }
        }
        EmbedCreateSpec embed = builder.build();
        Message message = channel.createMessage(embed).block();

        if (message != null && !ConfigEntry.DISCORD_REPORT_ARCHIVE_CHANNEL_ID.getString().isEmpty())
        {
            message.addReaction(ReactionEmoji.unicode("\uD83D\uDCCB")).subscribe();
        }

        return true;
    }

    @Override
    public boolean sendReport(Player reporter, Player reported, String reason)
    {
        if (!tfd4J.getBot().shouldISendReport())
        {
            return false;
        }

        final Guild server = tfd4J.getBot()
                .getClient()
                .getGuildById(SnowflakeEntry.serverID)
                .block();

        if (server == null)
        {
            FLog.severe("The guild ID specified in the config is invalid.");
            return false;
        }

        final TextChannel channel = server.getChannelById(SnowflakeEntry.reportChannelID)
                .ofType(TextChannel.class)
                .blockOptional()
                .orElseThrow();

        String location = "World: " + Objects.requireNonNull(reported.getLocation().getWorld()).getName() + ", X: " + reported.getLocation().getBlockX() + ", Y: " + reported.getLocation().getBlockY() + ", Z: " + reported.getLocation().getBlockZ();

        final EmbedCreateSpec spec = EmbedCreateSpec.builder()
                .title("Report for " + reported.getName())
                .description(reason)
                .footer("Reported by " + reporter.getName(), "https://minotar.net/helm/" + reporter.getName() + ".png")
                .timestamp(Instant.from(ZonedDateTime.now()))
                .addField("Location", location, true)
                .addField("Game Mode", WordUtils.capitalizeFully(reported.getGameMode().name()), true)
                .build();

        Message message = channel.createMessage(spec).block();

        if (!ConfigEntry.DISCORD_REPORT_ARCHIVE_CHANNEL_ID.getString().isEmpty() && message != null)
        {
            ReactionEmoji emoji = ReactionEmoji.unicode("\uD83D\uDCCB");
            message.addReaction(emoji);
        }

        return true;
    }

    @Override
    public boolean isEnabled()
    {
        return tfd4J.isEnabled();
    }
}
